/***************************************************************************
 * copyright            : (C) 2006 Andy Kelk <andy@mopoke.co.uk>           *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

 /**
  *  Based on njb mediadevice with some code hints from the libmtp
  *  example tools
  */

 /**
  *  MTP media device
  *  @author Andy Kelk <andy@mopoke.co.uk>
  *  @see http://libmtp.sourceforge.net/
  */

#define DEBUG_PREFIX "MtpMediaDevice"

#include <config.h>
#include "mtpmediadevice.h"

AMAROK_EXPORT_PLUGIN( MtpMediaDevice )

// Amarok
#include <debug.h>
#include <metabundle.h>
#include <statusbar/statusbar.h>
#include <statusbar/popupMessage.h>

// KDE
#include <kapplication.h>
#include <kiconloader.h>
#include <kpopupmenu.h>

// Qt
#include <qdir.h>
#include <qlistview.h>


/**
 * MtpMediaDevice Class
 */

MtpMediaDevice::MtpMediaDevice() : MediaDevice()
{
    m_name = "MTP Device";
    m_device = NULL;
    m_folders = NULL;
    setDisconnected();
    m_hasMountPoint = false;
    m_syncStats = false;
    m_transcode = false;
    m_transcodeAlways = false;
    m_transcodeRemove = false;
    m_configure = false;
    m_customButton = false;
    m_transfer = true;
}

void
MtpMediaDevice::init(MediaBrowser* parent)
{
    MediaDevice::init(parent);
}

bool
MtpMediaDevice::isConnected()
{
    return !(m_device == 0);
}

/**
 * File types that we support
 */
QStringList
MtpMediaDevice::supportedFiletypes()
{
    QStringList supportedFiles;
    supportedFiles << "mp3";
    supportedFiles << "wma";
    return supportedFiles;
}

/**
 * Copy a track to the device
 */
MediaItem
*MtpMediaDevice::copyTrackToDevice(const MetaBundle &bundle)
{
    DEBUG_BLOCK

    QString genericError = i18n( "Could not send track" );

    trackValueList::const_iterator it_track = m_trackList.findTrackByName( bundle.filename() );
    if( it_track != m_trackList.end() )
    {
        // track already exists. don't do anything (for now).
        debug() << "Track already exists on device." << endl;
        amaroK::StatusBar::instance()->shortLongMessage(
            genericError,
            i18n("Track already exists on device"),
            KDE::StatusBar::Error
        );
        return 0;
    }

    // read the mp3 header
    int duration = bundle.length();

    if( !duration )
    {
        debug() << "No duration" << endl;
        amaroK::StatusBar::instance()->shortLongMessage(
            genericError,
            i18n("Track has 0 duration"),
            KDE::StatusBar::Error
        );
        return 0;
    }

    LIBMTP_track_t* trackmeta = LIBMTP_new_track_t();
    trackmeta->item_id = 0;
    debug() << "filetype : " << bundle.fileType() << endl;
    if (bundle.fileType() == MetaBundle::mp3)
    {
        trackmeta->filetype = LIBMTP_FILETYPE_MP3;
    }
    else if (bundle.fileType() == MetaBundle::ogg)
    {
        trackmeta->filetype = LIBMTP_FILETYPE_OGG;
    }
    else if (bundle.fileType() == MetaBundle::wma)
    {
        trackmeta->filetype = LIBMTP_FILETYPE_WMA;
    }
    else if (bundle.fileType() == MetaBundle::mp4)
    {
        trackmeta->filetype = LIBMTP_FILETYPE_MP4;
    }
    else
    {
        debug() << "No filetype" << endl;
        amaroK::StatusBar::instance()->shortLongMessage(
            genericError,
            i18n("Cannot determine a valid file type"),
            KDE::StatusBar::Error
        );
        return 0;
    }

    if (bundle.title().isEmpty())
    {
        trackmeta->title = qstrdup("<Unknown>");
    }
    else
    {
        debug() << "Title: " << bundle.title() << endl;
        trackmeta->title = qstrdup(bundle.title().utf8());
    }

    if (bundle.album().isEmpty())
    {
        trackmeta->album = qstrdup("<Unknown>");
    }
    else
    {
        debug() << "Album: " << bundle.album() << endl;
        trackmeta->album = qstrdup(bundle.album().string().utf8());
    }

    if (bundle.artist().isEmpty())
    {
        trackmeta->artist = qstrdup("<Unknown>");
    }
    else
    {
        debug() << "Artist: " << bundle.artist() << endl;
        trackmeta->artist = qstrdup(bundle.artist().string().utf8());
    }

    if (bundle.genre().isEmpty())
    {
        trackmeta->genre = qstrdup("<Unknown>");
    }
    else
    {
        debug() << "Genre: " << bundle.genre() << endl;
        trackmeta->genre = qstrdup(bundle.genre().string().utf8());
    }

    if (bundle.year() > 0)
    {
        debug() << "Year: " << bundle.year() << endl;
        QString date;
        QTextOStream(&date) << bundle.year() << "0101T0000.0";
        trackmeta->date = qstrdup(date.utf8());
    }
    else
    {
        trackmeta->date = qstrdup("00010101T0000.0");
    }

    if (bundle.track() > 0) {
        debug() << "Track no: " << bundle.track() << endl;
        trackmeta->tracknumber = bundle.track();
    }
    if (bundle.length() > 0) {
        debug() << "Length: " << bundle.length() << endl;
        // Multiply by 1000 since this is in milliseconds
        trackmeta->duration = bundle.length() * 1000;
    }
    if (!bundle.filename().isEmpty())
    {
        trackmeta->filename = qstrdup(bundle.filename().utf8());
    }
    trackmeta->filesize = bundle.filesize();

    // Decide which folder to send it to:
    // If the device gave us a parent_folder setting, we use it
    uint32_t parent_id = 0;
    if (m_default_parent_folder >= 0)
    {
        parent_id = m_default_parent_folder;
        debug() << "Using default music folder : " << parent_id << endl;
    }
    // Otherwise look for a folder called "Music"
    else if (m_folders != NULL)
    {
        parent_id = folderNameToID("Music", m_folders, 0);
        if  (!parent_id)
        {
            debug() << "Parent folder could not be found. Going to use top level." << endl;
        }
    }
    // Give up and don't set a parent folder, let the device deal with it
    else
    {
        debug() << "No folders found. Going to use top level." << endl;
    }
    debug() << "Parent id : " << parent_id << endl;

    m_critical_mutex.lock();
    debug() <<"Sending track... " << bundle.url().path().utf8() << endl;
    int ret = LIBMTP_Send_Track_From_File(
        m_device, bundle.url().path().utf8(), trackmeta,
        NULL, NULL, parent_id
    );
    m_critical_mutex.unlock();

    if (ret < 0)
    {
        debug() << "Could not write file " << ret << endl;
        amaroK::StatusBar::instance()->shortLongMessage(
            genericError,
            i18n("File write failed"),
            KDE::StatusBar::Error
        );
        return 0;
    }

    MetaBundle temp( bundle );
    MtpTrack *taggedTrack = new MtpTrack( trackmeta );
    taggedTrack->setBundle( temp );

    LIBMTP_destroy_track_t(trackmeta);

    kapp->processEvents( 100 );

    // cache the track
    m_trackList.append( taggedTrack );
    return addTrackToView( taggedTrack );
}

/**
 * Recursively search the folder list for a matching one
 * and return its ID
 */
uint32_t
MtpMediaDevice::folderNameToID(char *name, LIBMTP_folder_t *folderlist, int level)
{
    uint32_t i;

    if(folderlist==NULL) {
        return 0;
    }

    if(!strcasecmp(name, folderlist->name))
        return folderlist->folder_id;

    if ((i = (folderNameToID(name, folderlist->child, level+1))))
        return i;
    if ((i = (folderNameToID(name, folderlist->sibling, level))))
        return i;

    return 0;
}

/**
 * Write any pending changes to the device, such as database changes
 */
void
MtpMediaDevice::synchronizeDevice()
{
    return;
}

/**
 * Find an existing track
 */
MediaItem *
MtpMediaDevice::trackExists( const MetaBundle& bundle )
{
    MediaItem *artist = dynamic_cast<MediaItem *>( m_view->findItem( bundle.artist(), 0 ) );
    if ( artist )
    {
        MediaItem *album = dynamic_cast<MediaItem *>( artist->findItem( bundle.album() ) );
        if( album )
        {
            return dynamic_cast<MediaItem *>( album->findItem( bundle.title() ) );
        }
    }
    return 0;
}

/**
 * Create a new playlist
 * @note Playlists not implemented yet... :-)
 */
MtpMediaItem *
MtpMediaDevice::newPlaylist(const QString &name, MediaItem *parent, QPtrList<MediaItem> items)
{
    Q_UNUSED(name);
    Q_UNUSED(parent);
    Q_UNUSED(items);
    return 0;
}

/**
 * Add an item to a playlist
 * @note Playlists not implemented yet... :-)
 */
void
MtpMediaDevice::addToPlaylist(MediaItem *mlist, MediaItem *after, QPtrList<MediaItem> items)
{
    Q_UNUSED(mlist);
    Q_UNUSED(after);
    Q_UNUSED(items);
}

/**
 * Recursively remove MediaItem from the tracklist and the device
 */
int
MtpMediaDevice::deleteItemFromDevice(MediaItem *mediaitem, bool onlyPlayed, bool deleteTrack )
{
    DEBUG_BLOCK

    int result = 0;

    MediaItem *next = 0;
    switch( mediaitem->type() )
    {
        case MediaItem::TRACK:
            if(mediaitem)
            {
                deleteMtpTrack( dynamic_cast<MtpMediaItem *> (mediaitem) );
                result++;
            }
            return result;
        case MediaItem::ALBUM:
        case MediaItem::ARTIST:
            // Recurse through the lists.
            for( MediaItem *it = dynamic_cast<MediaItem *>( mediaitem->firstChild() ); it ; it = next )
            {
                next = dynamic_cast<MediaItem *>(it->nextSibling());
                int res = deleteItemFromDevice( it, onlyPlayed, deleteTrack );
                if( res >= 0 && result >= 0 )
                    result += res;
                else
                    result = -1;
            }
            if(mediaitem)
                delete dynamic_cast<MediaItem *>(mediaitem);
            return result;
        default:
            return 0;
    }
    return result;
}

/**
 * Actually delete a track from the MTP Device
 */
int
MtpMediaDevice::deleteMtpTrack(MtpMediaItem *trackItem)
{
    DEBUG_BLOCK

    u_int32_t track_id = trackItem->track()->id();

    debug() << "delete this id : " << track_id << endl;

    QString genericError = i18n( "Could not delete track" );

    // delete the file
    m_critical_mutex.lock();
    int status = LIBMTP_Delete_Object(m_device, track_id);
    m_critical_mutex.unlock();
    if ( status != 0 )
    {
        debug() << "delete track failed" << endl;
        amaroK::StatusBar::instance()->shortLongMessage(
            genericError,
            i18n("Delete file failed"),
            KDE::StatusBar::Error
        );
        return -1;
    }
    debug() << "track deleted" << endl;

    // remove from the listview/tracklist
    m_trackList.remove(m_trackList.findTrackById( track_id ) );
    delete trackItem;
    kapp->processEvents( 100 );
    return 1;
}

/**
 * Connect to device, and populate m_view with MediaItems
 */
bool
MtpMediaDevice::openDevice( bool silent )
{
    DEBUG_BLOCK

    Q_UNUSED(silent);

    if( m_device != NULL )
        return true;


    QString genericError = i18n( "Could not connect to MTP Device" );

    m_critical_mutex.lock();
	LIBMTP_Init();
	m_device = LIBMTP_Get_First_Device();
    m_critical_mutex.unlock();
	if (m_device == NULL) {
        debug() << "No devices." << endl;
        amaroK::StatusBar::instance()->shortLongMessage( genericError, i18n("MTP device could not be opened"), KDE::StatusBar::Error );
        setDisconnected();
        return false;
	}

    QString modelname = QString(LIBMTP_Get_Modelname(m_device));
    QString ownername = QString(LIBMTP_Get_Ownername(m_device));
    m_name = modelname;
    if (! ownername.isEmpty())
    {
        m_name += " (" + ownername + ")";
    }

    m_default_parent_folder = m_device->default_music_folder;
    debug() << "setting default parent : " << m_default_parent_folder << endl;

    MtpMediaDevice::readMtpMusic();

    return true;
}

/**
 * Wrap up any loose ends and close the device
 */
bool
MtpMediaDevice::closeDevice()
{
    DEBUG_BLOCK

    // clear folder structure
    if (m_folders != NULL)
    {
        m_critical_mutex.lock();
        LIBMTP_destroy_folder_t(m_folders);
        m_critical_mutex.unlock();
        m_folders = NULL;
        debug() << "Folders destroyed" << endl;
    }

    // release device
    if (m_device != NULL)
    {
        m_critical_mutex.lock();
        LIBMTP_Release_Device(m_device);
        m_critical_mutex.unlock();
        setDisconnected();
        debug() << "Device released" << endl;
    }

    // clean up the view
    clearItems();

    return true;
}

/**
 * Get the capacity and freespace available on the device, in KB
 */
bool
MtpMediaDevice::getCapacity( KIO::filesize_t *total, KIO::filesize_t *available )
{
    DEBUG_BLOCK

    if (!isConnected())
        return false;

    uint64_t totalbytes;
    uint64_t freebytes;
    char *storage_description;
    char *volume_label;

    m_critical_mutex.lock();
    int ret = LIBMTP_Get_Storageinfo(m_device, &totalbytes, &freebytes, &storage_description, &volume_label);
    m_critical_mutex.unlock();
    if (ret == 0)
    {
        *total = totalbytes;
        *available = freebytes;
        return true;
    }
    else
    {
        debug() << "couldn't get storage details" << endl;
        return false;
    }
}

/**
 * Current device
 */
LIBMTP_mtpdevice_t*
MtpMediaDevice::current_device()
{
    return m_device;
}

/**
 * We use a NULL device to show a disconnected device.
 * This sets the device to that.
 */
void
MtpMediaDevice::setDisconnected()
{
    m_device = NULL;
}

/**
 * Handle clicking of the right mouse button
 */
void
MtpMediaDevice::rmbPressed(QListViewItem* qitem, const QPoint& point, int )
{

    enum Actions {DELETE};

    MtpMediaItem *item = static_cast<MtpMediaItem *>(qitem);
    if ( item )
    {
        KPopupMenu menu( m_view);
        menu.insertItem( SmallIconSet( amaroK::icon( "remove" ) ), i18n( "Delete from device" ), DELETE );

        int id =  menu.exec( point );
        switch( id )
        {
        case DELETE:
            MediaDevice::deleteFromDevice( item , false, true );
            break;
        }
        return;
    }

}

/**
 * Handle expanding a media item (show sub-items)
 */
void
MtpMediaDevice::expandItem( QListViewItem *item )
{
    DEBUG_BLOCK

    if (item == NULL)
        return;

    // First clear the item's children to repopulate.
    while( item->firstChild() )
        delete item->firstChild();

    MtpMediaItem *it = dynamic_cast<MtpMediaItem *>( item );

    switch( it->type() )
    {
        case MediaItem::ARTIST:
            if( it->childCount() == 0 ) // Just to be sure
                addAlbums( item->text( 0 ), it );
            break;
        case MediaItem::ALBUM:
            if( it->childCount() == 0 )
                addTracks( it->bundle()->artist(), item->text( 0 ), it );
            break;
        default:
            break;
    }
}

/**
 * Add a track to the current list view
 */
MtpMediaItem *
MtpMediaDevice::addTrackToView(MtpTrack *track, MtpMediaItem *item)
{
    QString artistName = track->bundle()->artist();

    MtpMediaItem *artist = dynamic_cast<MtpMediaItem *>( m_view->findItem( artistName, 0 ) );
    if(!artist)
    {
        artist = new MtpMediaItem(m_view);
        artist->m_device = this;
        artist->setText( 0, artistName );
        artist->setType( MediaItem::ARTIST );
    }

    QString albumName = track->bundle()->album();
    MtpMediaItem *album = dynamic_cast<MtpMediaItem *>( artist->findItem( albumName ) );
    if(!album)
    {
        album = new MtpMediaItem( artist );
        album->setText( 0, albumName );
        album->setType( MediaItem::ALBUM );
        album->m_device = this;
    }

    if( item )
        album->insertItem( item );
    else
    {
        item = new MtpMediaItem( album );
        item->m_device = this;
        QString titleName = track->bundle()->title();
        item->setTrack( track );

        item->setText( 0, titleName );
        item->setType( MediaItem::TRACK );
        item->setBundle( track->bundle() );
        item->track()->setId( track->id() );
    }
    return item;
}

/**
 * Add new albums to the tracklist
 */
MtpMediaItem *
MtpMediaDevice::addAlbums(const QString &artist, MtpMediaItem *item)
{
    for( trackValueList::iterator it = m_trackList.begin(); it != m_trackList.end(); it++ )
    {
        if( item->findItem( (*it)->bundle()->album() ) == 0 && ( (*it)->bundle()->artist().string() == artist ) )
        {
            MtpMediaItem *album = new MtpMediaItem( item );
            album->setText( 0, (*it)->bundle()->album() );
            album->setType( MediaItem::ALBUM );
            album->setExpandable( true );
            album->setBundle( (*it)->bundle() );
            album->m_device = this;
        }
    }
    return item;
}

/**
 * Add new tracks to the tracklist
 */
MtpMediaItem *
MtpMediaDevice::addTracks(const QString &artist, const QString &album, MtpMediaItem *item)
{
    for( trackValueList::iterator it = m_trackList.begin(); it != m_trackList.end(); it++ )
    {
        if( ( (*it)->bundle()->album().string() == album ) && ( (*it)->bundle()->artist().string() == artist ))
        {
            MtpMediaItem *track = new MtpMediaItem( item );
            track->setText( 0, (*it)->bundle()->title() );
            track->setType( MediaItem::TRACK );
            track->setBundle( (*it)->bundle() );
            track->setTrack( (*it) );
            track->m_device = this;
        }
    }
    return item;
}

/**
 * Get tracks and add them to the listview
 */
int
MtpMediaDevice::readMtpMusic()
{

    DEBUG_BLOCK

    int result = 0;

    if(m_trackList.isEmpty())
    {
        m_critical_mutex.lock();
        result = m_trackList.readFromDevice(this);
        m_critical_mutex.unlock();
    }

    debug()<< "Result : " << result << endl;

    clearItems();

    if(result == 0)
    {
        kapp->processEvents( 100 );

        for( trackValueList::iterator it = m_trackList.begin(); it != m_trackList.end(); it++ )
        {
            if( m_view->findItem( ((*it)->bundle()->artist().string()), 0 ) == 0 )
            {
                MtpMediaItem *artist = new MtpMediaItem( m_view );
                artist->setText( 0, (*it)->bundle()->artist() );
                artist->setType( MediaItem::ARTIST );
                artist->setExpandable( true );
                artist->setBundle( (*it)->bundle() );
                artist->m_device = this;
                expandItem(artist);
                expandItem(artist->firstChild());
            }
        }
    }
    return result;
}

/**
 * Clear the current listview
 */
void
MtpMediaDevice::clearItems()
{
    m_view->clear();
}

void
MtpMediaDevice::setFolders(LIBMTP_folder_t* folders)
{
    m_folders = folders;
}


/**
 * trackValueList Class
 */

/**
 * Find a track by its id
 */
trackValueList::iterator
trackValueList::findTrackById( unsigned _id )
{
    trackValueList::iterator it;
    for( it = begin(); it != end(); it++)
        if( (*it)->id() == _id)
            break;
    return it;
}

/**
 * Find a track by its id
 */
trackValueList::const_iterator
trackValueList::findTrackById( unsigned _id ) const
{
    trackValueList::const_iterator it;
    for( it = begin(); it != end(); it++)
        if( (*it)->id() == _id)
            break;
    return it;
}

/**
 * Find a track by its name
 */
trackValueList::iterator
trackValueList::findTrackByName( const QString& _filename )
{
    trackValueList::iterator it;
    for( it = begin(); it != end(); it++)
        if( (*it)->bundle()->url().path() == _filename)
            break;
    return it;
}

/**
 * Transfer info from the device to local structures
 */
int
trackValueList::readFromDevice( MtpMediaDevice* mtp )
{

    DEBUG_BLOCK

    LIBMTP_mtpdevice_t* device = mtp->current_device();

    QString genericError = i18n( "Could not get music from MTP Device" );

    LIBMTP_track_t* tracks = LIBMTP_Get_Tracklisting(device);

    debug() << "Got tracks from device" << endl;

    if (tracks == NULL)
    {
        debug()<< "Error reading tracks. NULL returned." << endl;
        amaroK::StatusBar::instance()->shortLongMessage( genericError, i18n("Could not read MTP Device tracks"), KDE::StatusBar::Error );
        return -1;
    }
    else
    {
        LIBMTP_track_t* tmp;
        while (tracks != NULL)
        {
            debug() << "Read track... : " << tracks << endl;
            MtpTrack* mtp_track = new MtpTrack( tracks );
            mtp_track->readMetaData( tracks );
            append( mtp_track );
            tmp = tracks;
            tracks = tracks->next;
            LIBMTP_destroy_track_t(tmp);
        }
        kapp->processEvents( 100 );
    }

    LIBMTP_folder_t* folders = LIBMTP_Get_Folder_List(device);
    mtp->setFolders(folders);

    return 0;
}

/**
 * MtpTrack Class
 */
MtpTrack::MtpTrack (LIBMTP_track_t* track)
{
    m_id = track->item_id;
}


MtpTrack::~MtpTrack()
{
    m_itemList.setAutoDelete( true );
    while( m_itemList.count() > 0 )
    {
        delete m_itemList.first();
    }
}

/**
 * Read track properties from the device and set it on the track
 */
void
MtpTrack::readMetaData(LIBMTP_track_t* track)
{
    MetaBundle* bundle = new MetaBundle();

    if (track->genre != NULL)
        bundle->setGenre( AtomicString(qstrdup(track->genre)) );
    if (track->artist != NULL)
        bundle->setArtist( AtomicString(qstrdup(track->artist)) );
    if (track->album != NULL)
        bundle->setAlbum( AtomicString(qstrdup(track->album)) );
    if (track->title != NULL)
        bundle->setTitle( AtomicString(qstrdup(track->title)) );

    // translate codecs to file types
    if ( track->filetype == LIBMTP_FILETYPE_MP3 )
        bundle->setFileType( MetaBundle::mp3 );
    else if ( track->filetype == LIBMTP_FILETYPE_WMA )
        bundle->setFileType( MetaBundle::wma );
    else if ( track->filetype == LIBMTP_FILETYPE_OGG )
        bundle->setFileType( MetaBundle::ogg );
    else
        bundle->setFileType( MetaBundle::other );

    if (track->date != NULL)
        bundle->setYear( QString(qstrdup(track->date)).mid(0,4).toUInt() );
    if (track->tracknumber > 0)
        bundle->setTrack( track->tracknumber );
    if (track->duration > 0)
        bundle->setLength( track->duration / 1000 ); // Divide by 1000 since this is in milliseconds

    this->setBundle( *bundle );
}

/**
 * Set this track's metabundle
 */
void
MtpTrack::setBundle( MetaBundle &bundle )
{
    m_bundle = bundle;
}

/**
 * Add a child item
 */
void
MtpTrack::addItem( const MtpMediaItem *item )
{
    m_itemList.append( item );
}

/**
 * Remove a child item
 */
bool
MtpTrack::removeItem( const MtpMediaItem *item )
{
    m_itemList.remove( item );
    return m_itemList.isEmpty();
}
